# powerline - Simple, easy, computer reservations system
# (c) 2008 Pianohacker, licensed under the GPLv3

from __future__ import absolute_import

import cherrypy
from os import path

class user_error(Exception):
	pass

from powerline import output, database, json

location = path.abspath(path.dirname(__file__))
output.template_dir = path.join(location, 'templates')

from powerline import manager, xmlrpc

cherrypy.config.namespaces.update(database = lambda k, v: setattr(database, k, v), output = lambda k, v: setattr(output, k, v))


__all__ = ['location', 'output', 'database', 'json', 'manager', 'xmlrpc']



# powerline.xmlrpc - Pre-book XML-RPC API
# (c) 2008 Pianohacker, licensed under the GPLv3

"""An implementation of the Pre-book XML-RPC API.

All Pre-book API methods take a single params argument, a dict, and return a dict.
While the system_api_procs.tcl file has five real methods, only four are actually used:
	* pcres.authenticate,
	* pcres.startsession,
	* pcres.infoupdate,
	* and pcres.closesession.

TODO: Consider a rewrite of the client and this API. The current model is ripe for abuse, and could deal with a bit of digest-authentication, or even something RESTful.
"""

import cherrypy
from cherrypy.lib import xmlrpc as _xmlrpc
from powerline import user_error
from powerline.output import join_list
from powerline.database import user_model, session_model, settings_model, system_model, hours_model, connect
from dbwrap import bag
import datetime

from xmlrpclib import Fault

class xmlrpc_controller(object):
    # This is custom because some part of the client-xmlrpclib chain puts in spurious spaces. Evilbadhorrible people
    # Note we're hard-coding this into the 'tools' namespace. We could do
    # a huge amount of work to make it relocatable, but the only reason why
    # would be if someone actually disabled the default_toolbox. Meh.
    _cp_config = {'tools.xmlrpc.on': True}
    
    def __call__(self, *vpath, **params):
        rpcparams, rpcmethod = _xmlrpc.process_body()
        
        subhandler = self
        for attr in str(rpcmethod).split('.'):
            subhandler = getattr(subhandler, attr.strip(), None)
         
        if subhandler and getattr(subhandler, "exposed", False):
            body = subhandler(*(vpath + rpcparams), **params)
        
        else:
            # http://www.cherrypy.org/ticket/533
            # if a method is not found, an xmlrpclib.Fault should be returned
            # raising an exception here will do that; see
            # cherrypy.lib.xmlrpc.on_error
            raise Exception, 'method "%s" is not supported' % attr
        
        conf = cherrypy.request.toolmaps['tools'].get("xmlrpc", {})
        _xmlrpc.respond(body,
                        conf.get('encoding', 'utf-8'),
                        conf.get('allow_none', 0))
        return cherrypy.response.body
    __call__.exposed = True
    
    index = __call__


def require_params(params, *args):
	"""Checks that the keys in the params dict contain all of args."""
	if not all(key in params for key in args):
		raise Fault(-32500, 'required params not sent (did not get %s)' % ', '.join(set(args) - set(params.keys())))

class api(xmlrpc_controller):
	_cp_config = {
		'tools.trailing_slash.on': False,
		'tools.xmlrpc.on': True
	}

	def __init__(self):
		self.pcres = real_api()

class real_api(api):
	"""The pcres namespace of the API."""
	def __init__(self):
		pass

	@cherrypy.expose
	def authenticate(self, params):
		"""Authenticates the user.

		This method takes four arguments inside its params dict:
			* auth_name: This is the barcode of the user.
			* auth_token: This is the user's password, if valid. Note that the official client sends '0' when no password is entered.
			* system_id: This is the internal name of the system.
			* station_id: This is the numerical id (counting from 0) of a sub-station on one physical system. This will be 0, in most cases. NOTE: Powerline does not support sub-stations.

		It returns a dict like so:
			* status_code: a short string, indicating the success of the call.
				* "SUCCESS": The user authenticated successfully.
				* "WALKUP_BOOKTIME": Undocumented and unused.
				* "WALKUP_LEFTTIME": Same.
				* "AUTHFAILED": The authentication step failed. The user does not exist, or we couldn't contact the external source, etc.
				* "BOOKERR": The patron is already logged on.
				* "INUSEERR": Undocumented and unused. Note that Powerline uses this to mean that the user is not close enough to the head of the line.
				* "SYSERR": An internal error occured.
			* status_message: A user-readable and understandable message to correspond to an error.
			* patron_type: The patron category code, if we are using ILS authentication.
			* session_max_seconds: The maximum length of the session in seconds.
			* reservation_id: An internal id for this reservation. NOTE: Powerline has a unified model of reservations and sessions, so this will be the same as the session_id.
		"""

		require_params(params, 'auth_name', 'auth_token', 'system_id', 'station_id')

		error = lambda status_code, status_message: dict(status_code = status_code, status_message = status_message, patron_type = 'failed', session_max_seconds = 0, reservation_id = 0)

		database = connect()
		user = user_model(database)
		system = system_model(database)
		session = session_model(database)
		settings = settings_model(database)
		hours = hours_model(database)

		session_max_seconds = min(settings['session_time'] * 60, (datetime.datetime.combine(datetime.date.today(), hours.get_closing_time()) - datetime.datetime.now()).seconds)

		if not hours.open_now():
			return error('SYSERR', 'Currently closed')

		if not system.exists(params['system_id']):
			return error('SYSERR', 'System is not registered')

		try:
			user.verify(params['auth_name'])
		except user_error:
			return error('AUTHFAILED', 'You are not registered')

		if params['auth_name'] in settings['overrides']:
			# Evilbadhorrible hackery
			session.create(params['auth_name'])
			user_session = session.get(params['auth_name'])
			session.authenticate(user_session.session_id, params['system_id'])
			return dict(status_code = 'SUCCESS', status_message = '', patron_type = 'OVERRIDE', session_max_seconds = session_max_seconds, reservation_id = user_session.session_id)

		if params['auth_name'] in settings['overrides']:
			return dict(status_code = 'SUCCESS', status_message = '', patron_type = 'ADULT', session_max_seconds = session_max_seconds, reservation_id = user_session.session_id)

		user_session = None

		try:
			user_session = session.get(params['auth_name'])

			if user_session.start_time is not None:
				return error('BOOKERR', 'You are already logged on')

		except KeyError:
			return error('AUTHFAILED', 'You have not reserved')

		open_computers = system.get_open()

		if open_computers == []:
			return error('SYSERR', 'Attempt to log on when all computers are full')

		if settings['session_expire_time'] != 0:
			first_in_line = session.select_one('start_time IS NULL LIMIT 1')
			last_ended = session.select_one('end_time IS NOT NULL AND end_time + INTERVAL 10 minute > CURRENT_TIMESTAMP AND DATE(timestamp) = CURRENT_DATE ORDER BY end_time DESC LIMIT 1')

			if first_in_line and last_ended and len(session.get_unstarted()) > len(open_computers) and last_ended.end_time + datetime.timedelta(minutes = settings['session_expire_time']) < datetime.datetime.now():
				if first_in_line.user == params['auth_name']:
					return error('INUSEERR', 'Your session has expired')

				session.delete(first_in_line.session_id)

		line_position = user.get_line_position(params['auth_name'])

		if line_position > len(open_computers):
			return error('INUSEERR', 'People are ahead of you in line')

		session.authenticate(user_session.session_id, params['system_id'])

		return dict(status_code = 'SUCCESS', status_message = '', patron_type = 'ADULT', session_max_seconds = session_max_seconds, reservation_id = user_session.session_id)

	@cherrypy.expose
	def startsession(self, params):
		"""Actually begin the session.

		While in Pre-book, this creates a session and returns it, in Powerline, it just adds a start_time to the reservation.

		This method takes two arguments in its params dict:
			* reservation_id: The previously given internal id for this reservation. See note in authenticate.
			* session_type: This appears to be similar to the User-Agent header in HTTP; e.g., the official client sends Win32Lock.
		
		It returns a dict like so:
			* status_code: Either "SUCCESS" (everything went well) or "SYSERR" (internal error).
			* session_id: An internal id for this session. In Powerline, same as the reservation_id.
		"""
		require_params(params, 'reservation_id', 'session_type')

		database = connect()
		session = session_model(database)

		result = dict(status_code = 'SUCCESS', session_id = 0, checkup_interval = 120)

		if not session.exists(session_id = params['reservation_id']):
			result['status_code'] = 'SYSERR'

			return result

		session.start(params['reservation_id'])

		result['session_id'] = params['reservation_id']

		return result

	@cherrypy.expose
	def closesession(self, params):
		"""Close the session.

		This method takes two arguments in its params dict:
		* session_id: The previously given internal id for this session.
		* session_end_code: A code giving the reason for the session end.
			* "user-logout": Manual logout before the session ended.
			* "privacy": Client logged out user because of inactivity.
			* "timeup": Normal logout at end of session.
			* "error": Client detected an error of some sort, and logged out.

		It returns a dict like so:
			* status_code: A short code indicating the success of the call.
				* "SUCCESS": Everything went well.
				* "BADSESSIONID": The session_id doesn't exist.
				* "SYSERR": An internal error occurred.
		"""
		require_params(params, 'session_id', 'session_end_code')

		database = connect()
		session = session_model(database)
		result = dict(status_code = 'SUCCESS')

		if not session.exists(session_id = params['session_id']):
			result['status_code'] = 'BADSESSIONID'

			return result

		session.end(params['session_id'])
	
		return result

	@cherrypy.expose
	def infoupdate(self, params):
		"""Get updated session info.

		This method takes one argument in its params dict:
			* session_id: The previously given internal id for this session.

		It returns a dict like so:
			* status_code: A short code indicating the success of the call.
				* "TIMEUNCHANGED": Nothing happened, session unaffected.
				* "TIMEEXTENDED": The client has been given more time.
				* "TIMEREDUCED": The client's session has been cut short.
				* "SYSERR": An internal error occurred.
			* status_message: A human-readable message for the client corresponding to the status_code, if something occurred.
			* session_remaining_seconds: Amount of time remaining in session, in seconds.

		Note that the official server software _always_ returns a dict like this:
			* status_code: "TIMEUNCHANGED"
			* status_message: "Your session time has been reduced 30 minutes by library staff"
			* session_remaining_seconds: 360

		These values do not correspond to anything in the real world, so this method can not be relied on for anything if compatibility with the official server software is required.
		"""
		require_params(params, 'session_id')

		database = connect()
		session = session_model(database)
		settings = settings_model(database)
		hours = hours_model(database)
		result = dict(status_code = 'TIMEUNCHANGED', status_message = '', session_remaining_seconds = 0)
		
		if not session.exists(session_id = params['session_id']):
			result['status_code'] = 'SYSERR'
			result['status_message'] = 'Bad session_id'
			result['session_remaining_seconds'] = 0
			
			return result

		now = datetime.datetime.now()

		user_session = session.get(session_id = params['session_id'])
		session_time = min(settings['session_time'] * 60, (datetime.datetime.combine(datetime.date.today(), hours.get_closing_time()) - datetime.datetime.now()).seconds)
		result['session_remaining_seconds'] = ((user_session.start_time + datetime.timedelta(0, session_time, 0)) - datetime.datetime.now()).seconds

		return result

	@cherrypy.expose
	def clearsession(self, params):
		"""Clear all sessions for a machine.

		This is supposed to be called when a computer reboots. Since the official client does not actually use this, implementing it is low priority.

		This method takes one argument in its params dict:
			* system_id: The internal id of the system.

		This method returns a dict like so:
			* status_code: Either "SUCCESS" (everything went well) or "SYSERR" (internal error).
		"""
		return dict(status_code = 'SUCCESS')



# powerline.manager - Management interface
# (c) 2008 Pianohacker, licensed under the GPLv3

import cherrypy
from powerline import output, user_error
from powerline.database import user_model, settings_model, system_model, session_model, hours_model, connect
from datetime import datetime

class user_controller(object):
	@output.expose(template = 'manager/users.html')
	def index(self, **kwargs):
		"""/manager/users/ - User management.

		GET - Returns the main page.
		POST - Creates a user (username parameter).
		"""
		database = connect()
		user = user_model(database)

		if cherrypy.request.method == 'POST':
			if user.exists(kwargs['username']):
				raise user_error('User %s already exists' % kwargs['username'])
			user.create(kwargs['username'])
			raise cherrypy.HTTPRedirect('/manager/users/')
		else:
			return {}

	@output.expose(template = 'manager/user.html')
	def default(self, user_id, **kwargs):
		"""/manager/users/<user_id> - Manage a particular user.

		GET - Returns details on a user.
		PUT - Creates a user.
		DELETE - Deletes a user.
		POST - Modifies a user.
		"""
		database = connect()
		user = user_model(database)
		method = ''

		if 'method' in kwargs:
			method = kwargs['method']
		else:
			method = cherrypy.request.method

		if method == 'GET':
			if user.exists(user_id):
				return {'result': user.get(user_id)}
			else:
				raise cherrypy.NotFound()
		elif method == 'PUT':
			if not user.exists(user_id):
				user.create(user_id)
			raise cherrypy.HTTPRedirect('/manager/users/' + user_id)
		elif method == 'DELETE':
			user.delete(user_id)
			raise cherrypy.HTTPRedirect('/manager/users/')
		elif method == 'POST':
			if not user.exists(user_id):
				raise cherrypy.NotFound()

			if 'barred' in kwargs:
				user.ban(user_id) if int(kwargs['barred']) else user.ban(user_id, '1970-01-01')
			elif 'ban_until' in kwargs:
				user.ban(user_id, datetime.strptime(kwargs['ban_until'], '%m/%d/%Y').strftime('%Y-%m-%d'))

			raise cherrypy.HTTPRedirect('/manager/users/' + user_id)

	@output.expose(template = 'manager/user-search.html')
	def search(self, query):
		"""/manager/users/search - Search names of users.

		Takes a single argument of `query`, which is searched for anywhere in the username of a user.
		"""
		database = connect()
		user = user_model(database)

		yield 'query', query
		yield 'result', user.select('username LIKE %s', '%' + query + '%')

class settings_controller(object):
	@output.expose(template = 'manager/settings.html')
	def index(self, key = '', value = '', type = '', description = ''):
		"""/manager/settings/ - Manage settings.

		POST - Takes arguments `key`, `value`, `type`, and `description`, and creates or modifies setting `name` accordingly.
		"""
		database = connect()
		settings = settings_model(database)

		if cherrypy.request.method == 'POST':
			if key in settings:
				settings[key] = value
			else:
				settings.insert(name = key, value = value, type = type, description = description)
				database.commit()

			raise cherrypy.HTTPRedirect('/manager/settings/')
		else:
			yield 'result', settings.all()
			yield 'types', settings.types.keys()

class hours_controller(object):
	@output.expose(template = 'manager/hours.html')
	def index(self, **kwargs):
		"""/manager/hours/ - Manage hours.

		GET - Returns current hours.
		POST - Modifies hours, based on arguments of the form <weekday>_(open|close)time, e.g., 0_opentime or 6_closetime.
		"""
		database = connect()
		hours = hours_model(database)
		current_hours = hours.all()

		if cherrypy.request.method == 'POST':
			for i in xrange(7):
				sent_properties = dict((key[2:], value) for key, value in kwargs.items() if key.startswith(str(i)))
				if sent_properties != {}:
					current_hours[i] = (datetime.strptime(sent_properties['opentime'], '%I:%M %p').time(), datetime.strptime(sent_properties['closetime'], '%I:%M %p').time())

			hours.update(current_hours)

			raise cherrypy.HTTPRedirect('/manager/hours/')
		else:
			return current_hours

class system_controller(object):
	"""Systems management"""
	@output.expose(template = 'manager/systems.html')
	def index(self, name = '', title = '', method = ''):
		"""/manager/systems/ - Manage systems.

		POST - Takes an argument of name and title, and creates a system with those properties.
		DELETE - Deletes the system with `name`.
		"""
		database = connect()
		system = system_model(database)

		if not method:
			method = cherrypy.request.method
	
		if method == 'POST':
			if not title:
				return 'Systems must be given a title'
			if system.exists(name):
				system.update(name, title = title)
			else:
				system.create(name = name, title = title)

			raise cherrypy.HTTPRedirect('/manager/systems/')
		elif method == 'DELETE':
			if system.exists(name):
				system.delete(name)

			raise cherrypy.HTTPRedirect('/manager/systems/')
		else:
			return system.all()

class interface(object):
	def __init__(self):
		self.users = user_controller()
		self.settings = settings_controller()
		self.systems = system_controller()
		self.hours = hours_controller()

		self._cp_config = {
			'tools.require_login.on': True
		}

	@output.expose(template = 'manager/index.html')
	def index(self):
		"""/manager/ - Main manager page"""
		return {}
	
	@output.expose(template = 'manager/login-setup.html')
	def login_setup(self, username = '', password = '', confirm_password = ''):
		"""/manager/login_setup - Set up the manager username and password.
		
		GET - Returns current username.
		POST - Modifies username and/or password, based on arguments.
		"""
		database = connect()
		settings = settings_model(database)

		yield 'username', settings['manager_users'][0].split(':')[0]

		if cherrypy.request.method == 'POST':
			assert(username and password)
			if password != confirm_password:
				raise user_error('Passwords do not match')

			settings['manager_users'] = username + ':' + password

			raise cherrypy.HTTPRedirect('/manager/')
	
	@output.expose(template = 'manager/session-log.html')
	def session_log(self, session_id = None, **kwargs):
		"""/manager/session_log - A log of sessions for the day"""
		database = connect()
		session = session_model(database)

		if 'method' in kwargs:
			method = kwargs['method']
		else:
			method = cherrypy.request.method

		if method == 'GET': 
			return session.select('date(timestamp) = current_date ORDER BY daily_id')
		elif method == 'DELETE':
			session_id = int(session_id)
			assert(session.exists(session_id = session_id))
			session.delete(session_id)

			raise cherrypy.HTTPRedirect('/manager/session_log')
	
	@output.expose(template = 'manager/login.html')
	def login(self, username = '', password = '', original_page = '/manager/'):
		"""/manager/login - Log in to the manager.

		GET - Returns the login page.
		POST - Checks username and password, and redirects to `original_page` if successful.
		"""
		database = connect()
		settings = settings_model(database)

		yield 'original_page', original_page

		if cherrypy.request.method == 'POST':
			if username + ':' + password in settings['manager_users']:
				cherrypy.session['logged-in'] = 'yes'
				yield 'destination', original_page
			else:
				raise user_error('Username or password incorrect')
	
	@cherrypy.expose
	def logout(self):
		"""/manager/logout - Log out from the manager"""
		if cherrypy.session.get('logged-in', False):
			del cherrypy.session['logged-in']

		raise cherrypy.HTTPRedirect('/')
	
	login._cp_config = {'tools.require_login.on': False}

	@output.expose(template = 'manager/statistics.html')
	def statistics(self):
		"""/manager/statistics - Returns statistics on sessions and users"""
		database = connect()
		session = session_model(database)
		user = user_model(database)

		return {
			'result': {
				 'sessions_for_day': session.rows('date(timestamp) = current_date').select_value('count(end_time)'),
				 'sessions_for_week': session.rows('week(timestamp) = week(current_date)').select_value('count(end_time)'),
				 'sessions_for_month': session.rows('month(timestamp) = month(current_date)').select_value('count(end_time)'),
				 'sessions_for_year': session.rows('year(timestamp) = year(current_date)').select_value('count(end_time)'),
				 'sessions': session.rows('1').select_value('count(end_time)'),
				 'registered_users': user.rows('1').select_value('count(*)'),
			}
		}



#!/usr/bin/env python

import pygtk
pygtk.require('2.0')
import gtk
import gobject
import pango
import xmlrpclib
import sys
import time

SERVER = 'http://localhost:8080/XMLRPC'
SYSTEMS = {'computer2': '#2', 'computer3': '#3', 'computer4': '#4', 'computer5': '#5'}
proxy, pcres = None, None

class user_error(Exception):
	pass

class client(object):
	session_end_code = 'user-logout'
	client_id = 'powerline test_client'

	def __init__(self, system_id):
		self.system_id = system_id
		self.reservation = {}
		self.session = {}
	
	def login(self, barcode):
		"""Logs in as 'barcode'.

		Usage: client.login('<barcode>')
		"""
		reservation = pcres.authenticate(dict(auth_name = barcode, auth_token = '', system_id = self.system_id, station_id = 0))
		if reservation['status_code'] != 'SUCCESS':
			raise user_error('%s in pcres.authenticate: %s' % (reservation['status_code'], reservation['status_message']))

		self.reservation = reservation

		session = pcres.startsession(dict(reservation_id = reservation['reservation_id'], session_type = self.client_id))
		if session['status_code'] != 'SUCCESS':
			raise user_error('%s in pcres.startsession: %s' % (session['status_code'], session['status_message']))

		self.session = session
	
	def update(self):
		"""Get updated session info.

		Usage: client.update()
		"""
		if not self.session:
			raise user_error('login first')

		return pcres.infoupdate(dict(session_id = self.session['session_id']))

	def logout(self):
		"""Logs out.

		Usage: client.logout()
		"""
		if not self.session:
			raise user_error('login first')

		result = pcres.closesession(dict(session_id = self.session['session_id'], session_end_code = self.session_end_code))
		if result['status_code'] != 'SUCCESS':
			raise user_error('%s in pcres.closesession' % result['status_code'])

class client_widget(gtk.VBox, client):
	def __init__(self, system_id):
		client.__init__(self, system_id)
		gtk.VBox.__init__(self, spacing = 20)
	
		self.status = gtk.Label('Please enter your barcode')
		status_attrs = pango.AttrList()
		status_attrs.insert(pango.AttrSize(12*1000, 0, 1000))
		self.status.set_attributes(status_attrs)
		self.pack_start(self.status)

		self.entry_box = gtk.HBox()
		self.pack_start(self.entry_box, expand = False)

		self.barcode_entry = gtk.Entry()
		self.barcode_entry.connect('activate', self.run_login)
		self.entry_box.pack_start(self.barcode_entry)

		self.barcode_submit = gtk.Button('Log in')
		self.barcode_submit.connect('clicked', self.run_login)
		self.entry_box.pack_start(self.barcode_submit)

		self.logout_button = gtk.Button('Log out')
		self.logout_button.set_no_show_all(True)
		self.logout_button.connect('clicked', self.run_logout)
		self.pack_start(self.logout_button, expand = False)

		self.show_all()

		self.session_time = 0
		self.updater = None
	
	def run_login(self, widget):
		try:
			self.login(self.barcode_entry.get_text())
		except user_error, e:
			dialog = gtk.MessageDialog(type = gtk.MESSAGE_ERROR, buttons = gtk.BUTTONS_OK, message_format = e.message)

			dialog.run(); dialog.destroy()

			return

		self.entry_box.hide()
		self.logout_button.show()

		self.session_time = int(self.reservation['session_max_seconds'])

		self.status.set_text('Logged in for %d minutes and %d seconds' % (
			self.session_time // 60,
			self.session_time % 60
		))

		self.updater = gobject.timeout_add(1000, self.run_update)

	def run_logout(self, widget):
		self.logout()

		self.status.set_text('Please enter your barcode')
		self.barcode_entry.set_text('')
		self.entry_box.show()
		self.logout_button.hide()

		gobject.source_remove(self.updater)
		self.session_time = 0
	
	def run_update(self):
		self.session_time -= 1

		if self.session_time == 0:
			self.run_logout()
			return False

		self.status.set_text('Logged in for %d minutes and %d seconds' % (
			self.session_time // 60,
			self.session_time % 60
		))

		return True

class client_window(gtk.Window):
	def destroy(self, widget, data=None):
		gtk.main_quit()

	def __init__(self):
		# create a new window
		gtk.Window.__init__(self)
	
		self.connect("destroy", self.destroy)
	
		self.container = gtk.Notebook()
		self.clients = []

		for id, title in SYSTEMS.items():
			client = client_widget(id)
			label = gtk.Label(title)
			label.set_padding(15, 0)
			self.clients.append(client)
			self.container.append_page(client, label)

		self.container.connect('switch-page', self.update_title)
		self.container.homogeneous = True
		self.add(self.container)

		self.set_default_size(400, 200)
		self.set_title('Powerline client - ' + SYSTEMS.itervalues().next())
		self.show_all()
	
	def update_title(self, notebook, page, page_num):
		title = self.container.get_tab_label_text(self.container.get_nth_page(page_num))

		self.set_title('Powerline client - ' + title)
		
def inputloop():
	my_client = client(sys.argv[1])

	while True:
		try:
			barcode = raw_input('Barcode > ')

			my_client.login(barcode)
			print 'Logged in for', '%d minutes and %d seconds' % (my_client.reservation['session_max_seconds'] // 60, my_client.reservation['session_max_seconds'] % 60), 'hit Ctrl-C to log out'
			while True:
				try:
					time.sleep(60)
					result = my_client.update()
					if result['status_code'] != 'TIMEUNCHANGED': print 'Session status: ', result['status_message']
					print 'Time left: ', '%d minutes and %d seconds' % (result['session_remaining_seconds'] // 60, result['session_remaining_seconds'] % 60)
				except KeyboardInterrupt:
					my_client.logout()
					break
		except user_error, e:
			print 'Error: ' + e.message
		except KeyboardInterrupt:
			return
		except:
			sys.__excepthook__(*sys.exc_info())

if __name__ == '__main__':
	if '--gtk' in sys.argv:
		if not 2 <= len(sys.argv) <= 3:
			print 'Usage: python -m powerline.tools.client [host]'
			print 'NOTE: host defaults to ' + SERVER
			raise SystemExit(1)

		if len(sys.argv) == 3:
			SERVER = sys.argv[2]

		proxy = xmlrpclib.ServerProxy(SERVER)
		print SERVER
		pcres = proxy.pcres

		main = client_window()
		try:
			gtk.main()
		finally:
			for client in main.clients:
				if client.session:
					try:
						client.logout()
					except:
						pass
	else:
		if not 2 <= len(sys.argv) <= 3:
			print 'Usage: python -m powerline.tools.client <system id> [host]'
			print 'NOTE: host defaults to ' + SERVER
			raise SystemExit(1)

		if len(sys.argv) == 3:
			SERVER = sys.argv[2]

		proxy = xmlrpclib.ServerProxy(SERVER)
		pcres = proxy.pcres

		print 'Powerline test client'
		inputloop()



# powerline.main - Main controllers and entry point
# (c) 2008 Pianohacker, licensed under the GPLv3

import cherrypy
from powerline import output, database, json, manager, xmlrpc, location, user_error
from powerline.lib import plugin
from powerline.database import user_model, system_model, session_model, settings_model, connect
import time
from sys import argv
from os import path

class root(object):
	data = cherrypy.tools.staticdir.handler(section = '/data', dir = 'templates/data')

	def __init__(self):
		self.XMLRPC = xmlrpc.api()
		self.manager = manager.interface()

		cherrypy.config.update({
			'tools.sessions.on': True,
			'tools.decode.on': True,
			'tools.encode.on': True,
			'tools.staticdir.root': location,
		})

		plugin.load(path.join(location, 'plugins'))

	@output.expose(template = 'index.html')
	def index(self, barcode = ''):
		"""/ - The main page. Also where reservation requests are POSTed to."""
		database = connect()
		user = user_model(database)
		session = session_model(database)
		system = system_model(database)
		settings = settings_model(database)

		yield 'reservation_entry_prefill', settings['reservation_entry_prefill']

		if cherrypy.session.get('logged-in', False):
			del cherrypy.session['logged-in']

		if not barcode:
			return

		try:
			sessions_per_user = settings['sessions_per_user']
			if session.rows('date(timestamp) = current_date AND user = %s', barcode).select_value('count(*)') >= sessions_per_user:
				raise user_error('You have already logged on %d times today' % sessions_per_user)
		except KeyError:
			pass

		session.create(barcode)
		reservation = session.get(barcode)
		reservation.update(dict(open_computers = system.get_open(), line_position = user.get_line_position(barcode), open_time = session.get_open_time()))

		yield 'result', reservation

	@output.expose()
	def sessions(self):
		"""/sessions - Returns the list of sessions."""
		database = connect()

		cherrypy.response.headers['Content-Type'] = 'application/json'

		session = session_model(database)
		user = user_model(database)
		system = system_model(database)

		results = session.get_unended()
		open_computers = system.get_open()

		for i, result in enumerate(results):
			result = results[i]
			result.line_position = user.get_line_position(result.user)
			if result.line_position != 0: result.line_position = output.get_ordinal(result.line_position)
			result.timestamp = time.mktime(result.timestamp.timetuple())
			if result.end_time: result.end_time = time.mktime(result.end_time.timetuple())
			if result.start_time: result.start_time = time.mktime(result.start_time.timetuple())
			del result.user
			results[i] = dict(result)

		return results
	
	@classmethod
	def start(klass, config_files):
		"""Starts powerline, and loads up each config file in `config_files`."""
		for config_file in config_files:
			cherrypy.config.update(config_file)

		cherrypy.quickstart(klass(), '/')


if __name__ == '__main__':
	if len(argv) == 1:
		print 'Usage: python -m powerline config_file [config_file...]'
		raise SystemExit(1)

	root.start(argv[1:])



from os import path, listdir
from cherrypy import log
import imp

class mount_point(type):
	known = set()

	def __init__(cls, name, bases, attrs):
		if hasattr(cls, '_plugins'):
			cls.add(cls)
		else:
			cls._plugins = set()
			if 'add' not in attrs:
				cls.add = classmethod(lambda cls, plugin: cls.plugins.add(plugin))
				cls.plugins = cls._plugins

			mount_point.known.add(cls)

def load(*paths):
	for path in paths:
		load_path(path)

def load_path(location):
	found = []
	for file in set([path.splitext(x)[0] for x in listdir(location) if not x.startswith('.')]):
		found += load_file(location, file)

	return found

def load_file(location, filename):
	found = []

	mod_name = path.splitext(filename)[0]
	new_mod = None

	try:
		fp, pathname, desc = imp.find_module(mod_name,[location])
		try:
			new_mod = imp.load_module(mod_name, fp, pathname, desc)
		finally:
			if fp: fp.close()
	except ImportError, err:
		log('Failed to import %s, %s' % (mod_name, err))

	if not new_mod:
		return []

	return load_module(new_mod)

def load_module(mod):
	if type(mod) is str:
		mod = _import(mod)

	objects = getattr(mod, "__all__", None) or [getattr(mod, x) for x in dir(mod)]

	return [obj for obj in objects if type(obj.__class__) in mount_point.known]



# powerline.database - Database access
# (c) 2008 Pianohacker, licensed under the GPLv3

import dbwrap
from os import path
from powerline.lib.plugin import mount_point
from powerline import user_error
import datetime
import MySQLdb

user = ''
host = 'localhost'
password = ''
database = 'powerline'
charset = 'utf8'

def connect():
	"""Returns a dbwrap'd connection to the MySQL database specified in config."""
	return dbwrap.wrapper(MySQLdb.connect(user = user, host = host, passwd = password, db = database, use_unicode = True, charset = charset), placeholder = '%s')

class user_checker(object):
	__metaclass__ = mount_point

class internal_checker(user_checker, dbwrap.table_ref):
	def __init__(self, connection):
		dbwrap.table_ref.__init__(self, connection, 'users')

	def verify(self, username = None):
		"""Returns the user with `username`, and adds a key 'banned' which indicates whether the user is banned."""
		user = self.rows(username = username).select_one('*', 'allowed_after >= CURRENT_DATE AS banned')
		
		if not user:
			raise KeyError(username)

		if user.banned:
			return False, 'You are banned'
		else:
			return True

	def exists(self, username):
		return self.rows(username = username).exist()

class user_model(dbwrap.table_ref):
	def __init__(self, connection):
		dbwrap.table_ref.__init__(self, connection, 'users')
		self.checkers = [checker(connection) for checker in user_checker.plugins]
	
	def create(self, username):
		self.insert(username = username)
		self.con.commit()
	
	def exists(self, username):
		return self.rows(username = username).exist()
	
	def verify(self, username):
		print 'verify("%s")' % username
		exists = False

		for checker in self.checkers:
			try:
				verified, message = checker.verify(username)
			except KeyError:
				continue

			exists = True

			if not verified:
				raise user_error(message)
		else:
			if not exists:
				raise user_error('User %s does not exist' % username)

	def get(self, username = None):
		"""Returns the user with `username`, and adds a key 'banned' which indicates whether the user is banned."""
		user = self.rows(username = username).select_one('*', 'allowed_after >= CURRENT_DATE AS banned')

		if not user:
			raise KeyError(user)

		return user
	
	def ban(self, username, until = datetime.date.max):
		"""Bans `username` until `until` if specified, otherwise forever."""
		self.rows(username = username).update(allowed_after = until)
		self.con.commit()
	
	def delete(self, username):
		assert(self.exists(username))

		self.rows(username = username).delete()
		self.con.commit()

	def get_line_position(self, username):
		"""Returns the position in line, counting from 1.

		If the user is already logged on, this returns -1. If they are ready to get on, it returns 0.
		"""
		session = session_model(self.con)
		system = system_model(self.con)

		try:
			reservation = session.get(username)

			if reservation.start_time != None:
				return -1
			
			for i, row in enumerate(session.get_unstarted()):
				if row.user == username:
					break
			else:
				raise user_error('fatal contradiction in get_line_position')

			return max(0, i + 1 - system.get_num_open())
		except KeyError:
			raise user_error('user does not have a reservation')
			
	__getitem__ = get


class system_model(dbwrap.table_ref):
	def __init__(self, connection):
		dbwrap.table_ref.__init__(self, connection, 'systems')
	
	def exists(self, system):
		return self.rows(name = system).exist()

	def create(self, name, title):
		self.insert(name = name, title = title)
		self.con.commit()
	
	def get(self, name):
		return self.select(name = name)

	def update(self, _name, **kwargs):
		"""Update the system `_name` to the values in kwargs."""
	 	super(system_model, self).rows(name = _name).update(**kwargs)
		self.con.commit()
	
	def delete(self, name):
		super(system_model, self).rows(name = name).delete()
		self.con.commit()
	
	def get_open(self):
		"""Returns the list of open systems."""
		return list(self.select('NOT EXISTS (SELECT session_id FROM sessions WHERE sessions.system = systems.name AND sessions.start_time IS NOT NULL AND sessions.end_time IS NULL)'))
	
	def get_num_open(self):
		"""Returns the number of open systems."""
		return self.rows('NOT EXISTS (SELECT session_id FROM sessions WHERE sessions.system = systems.name AND sessions.start_time IS NOT NULL AND sessions.end_time IS NULL)').select_value('COUNT(*)')
	
	def all_open(self):
		"""Returns whether all systems are open."""
		return len(self.all()) == self.get_num_open()

class session_model(dbwrap.table_ref):
	def __init__(self, connection):
		dbwrap.table_ref.__init__(self, connection, 'sessions')
		self.settings = settings_model(connection)
		self.users = user_model(connection)
		self.hours = hours_model(connection)
	
	def exists(self, username = None, **kwargs):
		return self.rows(**kwargs).exist() if kwargs else self.rows(user = username).exist()

	def create(self, username):
		"""Creates a session from the specified `username`."""

		# A little housekeeping...
		self.rows('date(timestamp) < current_date AND start_time IS NULL').delete()

		if not self.hours.open_now():
			raise user_error('Currently closed')

		self.users.verify(username)

		if self.rows(user = username, end_time = None).exist():
			raise user_error('You have already reserved a computer')
	
		last = self.rows('date(timestamp) = current_date').select_value('max(daily_id)')
		daily_id = 1
		if last is not None:
			daily_id = last + 1

		self.insert(user = username, daily_id = daily_id)
		self.con.commit()
	
	def delete(self, session_id):
		self.rows(session_id = session_id).delete()
		self.con.commit()

	def authenticate(self, session_id, system):
		"""Registers the specified `session_id` on `system`."""
		self.rows(session_id = session_id).update(system = system)
		self.con.commit()

	def start(self, session_id):
		"""Updates the start_time of of the session with `session_id` to the current time."""
		self.rows(session_id = session_id).update(start_time = datetime.datetime.now())
		self.con.commit()
	
	def end(self, session_id):
		"""Updates the end_time of of the session with `session_id` to the current time."""
		self.rows(session_id = session_id).update(end_time = datetime.datetime.now())
		self.con.commit()

	def get(self, username = None, **kwargs):
		"""If `username` is given, returns the first unended session for that user. Otherwise, it returns the session that matches `kwargs`."""
		results = self.select_one(**kwargs) if kwargs else self.select_one(user = username, end_time = None)

		if not results:
			raise KeyError(username or kwargs)
		else:
			return results

	__getitem__ = get

	def get_unstarted(self):
		"""Returns the list of unstarted sessions."""
		return self.select('start_time IS NULL AND date(timestamp) = current_date ORDER BY timestamp')

	def get_unended(self):
		"""Returns the list of unended sessions."""
		return self.select('end_time IS NULL AND date(timestamp) = current_date ORDER BY timestamp')

	def get_open_time(self):
		"""Returns when a system is likely to open up."""
		settings = settings_model(self.con)

		return self.rows('start_time IS NOT NULL ORDER BY start_time DESC LIMIT 1').select_value('start_time') + datetime.timedelta(minutes = settings['session_time'])


class settings_model(dbwrap.table_ref):
	"""A dict-like object that supports getting, setting and existence testing of settings.

	It also supports typed settings; if a setting is marked with a known type, then it is run through the function for that type.
	"""
	types = {
		'int': int,
		'float': float,
		'list': lambda val: [subval.strip() for subval in val.split(',')],
	}

	def __init__(self, connection):
		dbwrap.table_ref.__init__(self, connection, 'settings')
	
	def __getitem__(self, key):
		result = self.select_one(name = key)
		if result:
			if result.type in self.types:
				return self.types[result.type](result.value)
			else:
				return result.value
		else:
			raise KeyError(key)
	
	def __setitem__(self, key, value):
		if key in self:
			info = self.get_info(key)
			if info.type in self.types:
				self.types[info.type](value)

			self.rows(name = key).update(value = value)
		else:
			self.insert(name = key, value = value)

		self.con.commit()
	
	def __contains__(self, key):
		return self.rows(name = key).exist()
	
	def get_info(self, key):
		"""Returns the information for `key`, including its description and type."""
		result = self.select_one(name = key)
		
		if not result:
			raise KeyError(key)
		else:
			return result

class hours_model(object):
	"""A 'virtual' model, built on the 'hours' setting, that deals with hours."""
	def __init__(self, connection):
		self.settings = settings_model(connection)
	
	@staticmethod
	def _time_from_string(string):
		return datetime.datetime.strptime(string, '%H:%M').time()

	def all(self):
		"""Returns the hours for every day of the week, as a list of 2-tuples of datetime.time's."""
		return [tuple(self._time_from_string(time) for time in day.split('-')) for day in self.settings['hours']]
	
	def get(self, weekday = None):
		"""Returns the hours for `weekday` (specified numerically, starting from Sunday), if specified, otherwise today."""
		now = datetime.datetime.now()
		if weekday is None:
			weekday = now.date().weekday()
			weekday = 0 if weekday == 6 else weekday + 1

		return self.all()[weekday]

	def get_opening_time(self):
		"""Returns the opening time for today."""
		return self.get()[0]
	
	def get_closing_time(self):
		"""Returns the closing time for today."""
		return self.get()[1]

	def update(self, new_hours):
		"""Updates hours from `new_hours`, a list of 2-tuples of datetime.time's."""
		assert len(new_hours) == 7 and all(len(hours) == 2 and type(hours[0]) == type(hours[1]) == datetime.time for hours in new_hours)

		self.settings['hours'] = ','.join('-'.join(hour.strftime('%H:%M') for hour in hours) for hours in new_hours)
	
	def open_now(self):
		"""Returns whether we are currently open."""
		now = datetime.datetime.now().time()
		open, closing = self.get()

		return open <= now < closing



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
	name = 'powerline',
	version = '0.2.1',
	description = 'Simple computer reservations system',
	zip_safe = False,
	author = 'Pianohacker',
	author_email = 'pianohacker@gmail.com',
	license = 'GPLv3',
	url = 'http://code.google.com/p/powerline/',

	packages = find_packages(),
	include_package_data = True,
	package_data = {
		'powerline': [
			'templates/*.html',
			'templates/manager/*.html',
			'templates/data/*.js',
			'templates/data/*.css',
			'templates/data/images/*'
		]
	},
	zip_safe = False,

	install_requires = [
		'setuptools>=0.6b1',
		'Genshi>=0.4.1',
		'CherryPy>=3.0.0',
		'dbwrap>=0.1.6',
		'MySQL-python'
	],
)



# powerline.output - Output and formatting functions
# (c) 2008 Pianohacker, licensed under the GPLv3

from functools import wraps
import types
from os import path
from genshi.template.loader import TemplateLoader
from powerline import json as jsonlib, user_error
import cherrypy

"""Output and formatting functions.

Functions
=========
expose is a decorator that replaces CherryPy's @expose and the x.exposed = True idiom. It renders output to a template, and performs JSON serialization of results and errors.

get_ordinal and join_list are both for string formatting. Both currently only work with English, unfortunately.

Configuration Globals
=====================

template_dir (output.template_dir): Where the templates are loaded from.
"""

template_dir = ''

def expose(template = None, method = 'xhtml', **render_kwargs):
	"""A decorator that renders the return value of its wrapped function in a template, or serializes it to a template..

	Templates are loaded from the templates directory in the directory of the module.

	The return value of the function is transformed into a variables dict for the template in the following ways:
		* If the function is a generator, it is expected to yield 'key', 'value' tuples to be fed into dict().
		* If the value is a string, it is used unmodified.
		* If the value is a dict, it is used verbatim.
		* Anything else is an error.
		* If 'destination' is in the result dict, a redirect is fired with that value.
		* `method` and **render_kwargs are both passed to Genshi's render.

	If this is an XMLHttpRequest, it is serialized to JSON as follows:
		* A dict is directly serialized, with 'success': True added if not present.
		* Anything else is placed into a dict like {'result': result, 'success': True}
		* If an error occurs, it is serialized to {'success': False, 'error_type': type(e).__name__, 'error': str(e).
	"""

	def decorator(func):
		loader = None
		if not path.exists(template_dir):
			cherrypy.log('Template directory not valid')
			raise SystemExit(1)

		loader = TemplateLoader(template_dir, auto_reload = True)

		@wraps(func)
		def wrapper(*args, **kwargs):
			try: 
				try:
					body = {}
					result = func(*args, **kwargs)

					if type(result) is types.GeneratorType:
						body.update(result)
					else:
						body = result
				except user_error, e:
					body.update({
						'error_type': 'user_error',
						'error': str(e),
						'success': False,
					})

				if cherrypy.request.headers.get('X-Requested-With') == 'XMLHttpRequest':
					if isinstance(body, dict):
						if 'destination' in body:
							del body['destination']

						body.setdefault('success', True)
						return jsonlib.write(body)
					else:
						return jsonlib.write({'success': True, 'result': body})

				if not isinstance(body, (dict, basestring)):
					body = {'result': body, 'success': True}

				if isinstance(body, basestring):
					return result
				elif isinstance(body, dict):
					if 'destination' in body:
						raise cherrypy.HTTPRedirect(body['destination'])

					if template is None:
						return loader.load('dump.html').generate(result = body).render(method, **render_kwargs)
					else:
						return loader.load(template).generate(**body).render(method, **render_kwargs)
			except Exception, e:
				if cherrypy.request.headers.get('X-Requested-With') == 'XMLHttpRequest':
					if type(e) is cherrypy.HTTPRedirect:
						return {'success': True}

					if isinstance(e, cherrypy.HTTPError):
						cherrypy.response.status = e.status, ''

					return jsonlib.write({'success': False, 'error_type': type(e).__name__, 'error': str(e)})
				else:
					raise

		wrapper.exposed = True

		return wrapper
	
	return decorator

def require_login(login_page = '/manager/login'):
	"""A tool that requires a logged-in session for access; `login_page` is where it redirects to."""
	if not cherrypy.session.get('logged-in', False):
		raise cherrypy.HTTPRedirect(login_page + '?original_page=' + cherrypy.request.app.script_name + cherrypy.request.path_info)

cherrypy.tools.require_login = cherrypy.Tool('before_handler', require_login, priority = 70)

def get_ordinal(num):
	# FIXME: Port this to PyICU
	# Copied from http://vandahm.org/2006/10/03/cool-python-function/
	"""Transforms a number into its ordinal equivalent.

	>>> get_ordinal(1), get_ordinal(2), get_ordinal(3), get_ordinal(4)
	('1st', '2nd', '3rd', '4th')
	>>> get_ordinal(11), get_ordinal(21), get_ordinal(50002)
	('11th', '21st', '50002nd')
	"""
	
	special_suffixes = { '1': 'st', '2': 'nd', '3': 'rd' }
	default_return = 'th'

	digits = str(abs(num)) # To work with negative numbers
	last_digit = digits[-1:]

	if last_digit in special_suffixes.keys():
		# Numbers that end in 11, 12, and 13 just get 'th'
		if len(digits) == 1 or digits[-2] != '1':
			default_return = special_suffixes[last_digit]

	return str(num) + default_return

def join_list(list_, conjunction = 'or'):
	#FIXME: Port this to PyICU if possible
	"""Joins a list together in a readable form.

	>>> join_list(['a', 'b', 'c'])
	'a, b or c'
	>>> join_list(['a', 'b'])
	'a or b'
	>>> join_list(['a', 'b', 'c'], 'and')
	'a, b and c'
	>>> join_list(['a'])
	'a'
	>>> join_list([])
	''
	"""

	if len(list_) == 0:
		return ''
	elif len(list_) == 1:
		return str(list_[0])
	elif len(list_) == 2:
		return (' %s ' % conjunction).join(list_)
	else:
		return ', '.join(list_[:-1]) + (' %s ' % conjunction) + list_[-1]

if __name__ == '__main__':
	import doctest; doctest.testmod()



import string
import types
import datetime

##	json.py implements a JSON (http://json.org) reader and writer.
##	Copyright (C) 2005  Patrick D. Logan
##	Contact mailto:patrickdlogan@stardecisions.com
##
##	This library is free software; you can redistribute it and/or
##	modify it under the terms of the GNU Lesser General Public
##	License as published by the Free Software Foundation; either
##	version 2.1 of the License, or (at your option) any later version.
##
##	This library is distributed in the hope that it will be useful,
##	but WITHOUT ANY WARRANTY; without even the implied warranty of
##	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##	Lesser General Public License for more details.
##
##	You should have received a copy of the GNU Lesser General Public
##	License along with this library; if not, write to the Free Software
##	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


class _StringGenerator(object):
	def __init__(self, string):
		self.string = string
		self.index = -1
	def peek(self):
		i = self.index + 1
		if i < len(self.string):
			return self.string[i]
		else:
			return None
	def next(self):
		self.index += 1
		if self.index < len(self.string):
			return self.string[self.index]
		else:
			raise StopIteration
	def all(self):
		return self.string

class WriteException(Exception):
	pass

class ReadException(Exception):
	pass

class JsonReader(object):
	hex_digits = {'A': 10,'B': 11,'C': 12,'D': 13,'E': 14,'F':15}
	escapes = {'t':'\t','n':'\n','f':'\f','r':'\r','b':'\b'}

	def read(self, s):
		self._generator = _StringGenerator(s)
		result = self._read()
		return result

	def _read(self):
		self._eatWhitespace()
		peek = self._peek()
		if peek is None:
			raise ReadException, "Nothing to read: '%s'" % self._generator.all()
		if peek == '{':
			return self._readObject()
		elif peek == '[':
			return self._readArray()			
		elif peek == '"':
			return self._readString()
		elif peek == '-' or peek.isdigit():
			return self._readNumber()
		elif peek == 't':
			return self._readTrue()
		elif peek == 'f':
			return self._readFalse()
		elif peek == 'n':
			return self._readNull()
		elif peek == '/':
			self._readComment()
			return self._read()
		else:
			raise ReadException, "Input is not valid JSON: '%s'" % self._generator.all()

	def _readTrue(self):
		self._assertNext('t', "true")
		self._assertNext('r', "true")
		self._assertNext('u', "true")
		self._assertNext('e', "true")
		return True

	def _readFalse(self):
		self._assertNext('f', "false")
		self._assertNext('a', "false")
		self._assertNext('l', "false")
		self._assertNext('s', "false")
		self._assertNext('e', "false")
		return False

	def _readNull(self):
		self._assertNext('n', "null")
		self._assertNext('u', "null")
		self._assertNext('l', "null")
		self._assertNext('l', "null")
		return None

	def _assertNext(self, ch, target):
		if self._next() != ch:
			raise ReadException, "Trying to read %s: '%s'" % (target, self._generator.all())

	def _readNumber(self):
		isfloat = False
		result = self._next()
		peek = self._peek()
		while peek is not None and (peek.isdigit() or peek == "."):
			isfloat = isfloat or peek == "."
			result = result + self._next()
			peek = self._peek()
		try:
			if isfloat:
				return float(result)
			else:
				return int(result)
		except ValueError:
			raise ReadException, "Not a valid JSON number: '%s'" % result

	def _readString(self):
		result = ""
		assert self._next() == '"'
		try:
			while self._peek() != '"':
				ch = self._next()
				if ch == "\\":
					ch = self._next()
					if ch in 'brnft':
						ch = self.escapes[ch]
					elif ch == "u":
						ch4096 = self._next()
						ch256  = self._next()
						ch16   = self._next()
						ch1	= self._next()
						n = 4096 * self._hexDigitToInt(ch4096)
						n += 256 * self._hexDigitToInt(ch256)
						n += 16  * self._hexDigitToInt(ch16)
						n += self._hexDigitToInt(ch1)
						ch = unichr(n)
					elif ch not in '"/\\':
						raise ReadException, "Not a valid escaped JSON character: '%s' in %s" % (ch, self._generator.all())
				result = result + ch
		except StopIteration:
			raise ReadException, "Not a valid JSON string: '%s'" % self._generator.all()
		assert self._next() == '"'
		return result

	def _hexDigitToInt(self, ch):
		try:
			result = self.hex_digits[ch.upper()]
		except KeyError:
			try:
				result = int(ch)
			except ValueError:
				 raise ReadException, "The character %s is not a hex digit." % ch
		return result

	def _readComment(self):
		assert self._next() == "/"
		second = self._next()
		if second == "/":
			self._readDoubleSolidusComment()
		elif second == '*':
			self._readCStyleComment()
		else:
			raise ReadException, "Not a valid JSON comment: %s" % self._generator.all()

	def _readCStyleComment(self):
		try:
			done = False
			while not done:
				ch = self._next()
				done = (ch == "*" and self._peek() == "/")
				if not done and ch == "/" and self._peek() == "*":
					raise ReadException, "Not a valid JSON comment: %s, '/*' cannot be embedded in the comment." % self._generator.all()
			self._next()
		except StopIteration:
			raise ReadException, "Not a valid JSON comment: %s, expected */" % self._generator.all()

	def _readDoubleSolidusComment(self):
		try:
			ch = self._next()
			while ch != "\r" and ch != "\n":
				ch = self._next()
		except StopIteration:
			pass

	def _readArray(self):
		result = []
		assert self._next() == '['
		done = self._peek() == ']'
		while not done:
			item = self._read()
			result.append(item)
			self._eatWhitespace()
			done = self._peek() == ']'
			if not done:
				ch = self._next()
				if ch != ",":
					raise ReadException, "Not a valid JSON array: '%s' due to: '%s'" % (self._generator.all(), ch)
		assert ']' == self._next()
		return result

	def _readObject(self):
		result = {}
		assert self._next() == '{'
		done = self._peek() == '}'
		while not done:
			key = self._read()
			if type(key) is not types.StringType:
				raise ReadException, "Not a valid JSON object key (should be a string): %s" % key
			self._eatWhitespace()
			ch = self._next()
			if ch != ":":
				raise ReadException, "Not a valid JSON object: '%s' due to: '%s'" % (self._generator.all(), ch)
			self._eatWhitespace()
			val = self._read()
			result[key] = val
			self._eatWhitespace()
			done = self._peek() == '}'
			if not done:
				ch = self._next()
				if ch != ",":
					raise ReadException, "Not a valid JSON array: '%s' due to: '%s'" % (self._generator.all(), ch)
			assert self._next() == "}"
		return result

	def _eatWhitespace(self):
		p = self._peek()
		while p is not None and p in string.whitespace or p == '/':
			if p == '/':
				self._readComment()
			else:
				self._next()
			p = self._peek()

	def _peek(self):
		return self._generator.peek()

	def _next(self):
		return self._generator.next()

class JsonWriter(object):
		
	def _append(self, s):
		self._results.append(s)

	def write(self, obj, escaped_forward_slash=False):
		self._escaped_forward_slash = escaped_forward_slash
		self._results = []
		self._write(obj)
		return "".join(self._results)

	def _write(self, obj):
		if type(obj) in (datetime.time, datetime.date, datetime.datetime):
			obj = str(obj)
		
		ty = type(obj)
		if isinstance(obj, dict):
			n = len(obj)
			self._append("{")
			for k, v in obj.items():
				self._write(k)
				self._append(":")
				self._write(v)
				n = n - 1
				if n > 0:
					self._append(",")
			self._append("}")
		elif ty is types.ListType or ty is types.TupleType:
			n = len(obj)
			self._append("[")
			for item in obj:
				self._write(item)
				n = n - 1
				if n > 0:
					self._append(",")
			self._append("]")
		elif ty is types.StringType or ty is types.UnicodeType:
			self._append('"')
			obj = obj.replace('\\', r'\\')
			if self._escaped_forward_slash:
				obj = obj.replace('/', r'\/')
			obj = obj.replace('"', r'\"')
			obj = obj.replace('\b', r'\b')
			obj = obj.replace('\f', r'\f')
			obj = obj.replace('\n', r'\n')
			obj = obj.replace('\r', r'\r')
			obj = obj.replace('\t', r'\t')
			self._append(obj)
			self._append('"')
		elif ty is types.IntType or ty is types.LongType:
			self._append(str(obj))
		elif ty is types.FloatType:
			self._append("%f" % obj)
		elif obj is True:
			self._append("true")
		elif obj is False:
			self._append("false")
		elif obj is None:
			self._append("null")
		else:
			raise WriteException, "Cannot write in JSON: %s" % repr(obj)

def write(obj, escaped_forward_slash=False):
	return JsonWriter().write(obj, escaped_forward_slash)

def read(s):
	return JsonReader().read(s)



#!/usr/bin/python
# powerline.tools.wsgi_server
# This is a script for mod_wsgi.

CONFIGURATION_FILES = ['/opt/powerline/etc/powerline.conf']

if __name__ == '__main__':
	import sys
	sys.stdout = sys.stderr

	import atexit
	import cherrypy

	from powerline.main import root

	for config_file in CONFIGURATION_FILES:
		cherrypy.config.update(config_file)

	cherrypy.config.update({'environment': 'embedded'})

	if cherrypy.engine.state == 0:
		cherrypy.engine.start(blocking=False)
		atexit.register(cherrypy.engine.stop)

	application = cherrypy.Application(root(), None)
